using EPiServer.Commerce.Order;
using Mediachase.Commerce.Orders;
using Mediachase.Commerce.WorkflowCompatibility;
using Mediachase.MetaDataPlus;
using System.Linq;

namespace Mediachase.Commerce.Workflow.Activities
{
    public class CalculateTotalsActivity : OrderGroupActivityBase
    {
        private Currency _billingCurrency;

        /// <summary>
        /// Called by the workflow runtime to execute an activity.
        /// </summary>
        /// <param name="executionContext">The <see cref="T:Mediachase.Commerce.WorkflowCompatibility.ActivityExecutionContext"/> to associate with this <see cref="T:Mediachase.Commerce.WorkflowCompatibility.Activity"/> and execution.</param>
        /// <returns>
        /// The <see cref="T:Mediachase.Commerce.WorkflowCompatibility.ActivityExecutionStatus"/> of the run task, which determines whether the activity remains in the executing state, or transitions to the closed state.
        /// </returns>
        protected override ActivityExecutionStatus Execute(ActivityExecutionContext executionContext)
        {
            // Validate the properties at runtime
            ValidateRuntime();

            // Calculate order totals
            CalculateTotals();

            // Retun the closed status indicating that this activity is complete.
            return ActivityExecutionStatus.Closed;
        }

        /// <summary>
        /// Calculates the totals.
        /// </summary>
        private void CalculateTotals()
        {
            if (OrderGroup.ObjectState == MetaObjectState.Deleted)
            {
                return;
            }

            decimal subTotal = 0m;
            decimal shippingTotal = 0m;
            decimal handlingTotal = 0m;
            decimal taxTotal = 0m;
            decimal total = 0m;
            decimal orderDiscountAmount = 0m;

            // Get the property, since it is expensive process, make sure to get it once
            OrderGroup order = OrderGroup;
            _billingCurrency = order.BillingCurrency;

            // Calculate totals for OrderForms
            foreach (OrderForm form in order.OrderForms)
            {
                // Calculate totals for order form
                CalculateTotalsOrderForms(form);

                subTotal += form.SubTotal;
                shippingTotal += form.ShippingTotal;
                handlingTotal += form.HandlingTotal;
                taxTotal += form.TaxTotal;

                orderDiscountAmount += form.LineItems.Where(x => x.ObjectState != MetaObjectState.Deleted).Sum(li => li.OrderLevelDiscountAmount);
                total += form.Total;
            }

            // calculate OrderGroup totals
            order.SubTotal = subTotal;
            order.ShippingTotal = shippingTotal;
            order.TaxTotal = taxTotal;
            order.Total = total - orderDiscountAmount;
            order.HandlingTotal = handlingTotal;
        }

        /// <summary>
        /// Calculates the totals order forms.
        /// </summary>
        /// <param name="form">The form.</param>
        private void CalculateTotalsOrderForms(OrderForm form)
        {
            decimal subTotal = 0m;
            decimal discountTotal = 0m;
            decimal shippingDiscountTotal = 0m;
            decimal shippingTotal = 0m;

            foreach (LineItem item in form.LineItems.Where(x => x.ObjectState != MetaObjectState.Deleted))
            {
                // We should not include OrderLevelDiscountAmount in calculating SubTotal/Total of OrderForm and SubTotal of Shipment
                // the OrderLevelDiscountAmount will be included only in calculating Total of OrderGroup. It's the same behaviour as calculators.
                var itemDiscountedPrice = GetDiscountedPrice(item);

                item.ExtendedPrice = itemDiscountedPrice - item.OrderLevelDiscountAmount;
                subTotal += itemDiscountedPrice;
                discountTotal += item.LineItemDiscountAmount + item.OrderLevelDiscountAmount;
            }

            foreach (Shipment shipment in form.Shipments.Where(x => x.ObjectState != MetaObjectState.Deleted))
            {
                shipment.SubTotal = CalculateShipmentSubtotal(shipment);
                shippingTotal += shipment.ShippingSubTotal - shipment.ShippingDiscountAmount;
                shippingDiscountTotal += shipment.ShippingDiscountAmount;
            }

            form.ShippingTotal = shippingTotal;
            form.DiscountAmount = discountTotal + shippingDiscountTotal;
            form.SubTotal = subTotal;
            form.Total = subTotal + shippingTotal + form.TaxTotal;

            form.UpdatePaymentTotals();
        }

        private decimal CalculateShipmentSubtotal(Shipment shipment)
        {
            var lineItems = Shipment.GetShipmentLineItems(shipment).Where(item => item.Quantity > 0);
            return lineItems.Any() ? lineItems.Sum(item => GetItemPriceInShipment(item, shipment)) : 0m;
        }

        private decimal GetDiscountedPrice(LineItem lineItem)
        {
            return lineItem.PlacedPrice * lineItem.Quantity - lineItem.LineItemDiscountAmount;
        }

        private decimal GetItemPriceInShipment(LineItem lineItem, Shipment shipment)
        {
            return _billingCurrency.Round(GetDiscountedPrice(lineItem) / lineItem.Quantity * Shipment.GetLineItemQuantity(shipment, lineItem.LineItemId));
        }
    }
}