﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using System.Linq;
using System.Text;
using System.Web;
using System.Web.Caching;
using System.Xml.Serialization;
using EPiServer.ServiceLocation;
using Mediachase.Commerce.Catalog;
using Mediachase.Commerce.Catalog.Dto;
using Mediachase.Commerce.Catalog.Managers;
using Mediachase.Commerce.Catalog.Objects;
using Mediachase.Commerce.Core;
using Mediachase.Commerce.Website.BaseControls;
using Mediachase.Commerce.Website.Helpers;
using Mediachase.Data.Provider;
using Mediachase.Search;
using Mediachase.Search.Extensions;
using AppContext = Mediachase.Commerce.Core.AppContext;

namespace Mediachase.Commerce.Website.Search
{
    public class SearchFilterHelper
    {
        #region Local Fields
        private static object _LockObject = new object();
        private static volatile SearchConfig _SearchConfig = null;
        private SearchConfig _DynamicSearchConfig = null;
        private SearchManager _SearchManager = null;
        private static volatile Dictionary<string, CacheDependency> _Dependencies = new Dictionary<string, System.Web.Caching.CacheDependency>();
        private NameValueCollection _queryString = null;
        #endregion

        private static Injected<NodeIdentityResolver> NodeIdentityResolver;

        /// <summary>
        /// Gets the manager.
        /// </summary>
        /// <value>The manager.</value>
        private SearchManager Manager
        {
            get
            {
                if (_SearchManager == null)
                {
                    _SearchManager = new SearchManager(AppContext.Current.ApplicationName);
                }

                return _SearchManager;
            }
        }

        /// <summary>
        /// Gets the current URL.
        /// </summary>
        /// <value>The current URL.</value>
        private static string CurrentUrl
        {
            get
            {
                return HttpContext.Current.Request.RawUrl;
            }
        }

        #region Public Properties
        /// <summary>
        /// Gets the current.
        /// </summary>
        /// <value>The current.</value>
        public static SearchFilterHelper Current
        {
            get
            {
                SearchFilterHelper helper = null;
                if (HttpContext.Current.Items["Mediachase.SearchFilterHelper"] == null)
                {
                    helper = new SearchFilterHelper();
                    HttpContext.Current.Items["Mediachase.SearchFilterHelper"] = helper;
                }
                else
                {
                    helper = (SearchFilterHelper)HttpContext.Current.Items["Mediachase.SearchFilterHelper"];
                }

                return helper;
            }
        }

        private ISearchResults _Results;

        /// <summary>
        /// Gets the results.
        /// </summary>
        /// <value>The results.</value>
        public ISearchResults Results
        {
            get { return _Results; }
        }

        private string _SearchConfigPath = "~/Configs/Mediachase.Search.Filters.config";

        /// <summary>
        /// Gets or sets the search config path.
        /// </summary>
        /// <value>The search config path.</value>
        public string SearchConfigPath
        {
            get { return _SearchConfigPath; }
            set { _SearchConfigPath = value; }
        }


        /// <summary>
        /// Gets the search config.
        /// </summary>
        /// <value>The search config.</value>
        /// <returns></returns>
        public SearchConfig SearchConfig
        {
            get
            {
                bool changed = HasConfigChanged("BaseConfig");
                if (_SearchConfig == null || changed)
                {
                    lock (_LockObject)
                    {
                        changed = HasConfigChanged("BaseConfig");
                        if (_SearchConfig == null || changed)
                        {
                            string path = HttpContext.Current.Server.MapPath(SearchConfigPath);
                            TextReader reader = new StreamReader(path);
                            XmlSerializer serializer = new XmlSerializer(typeof(SearchConfig));
                            _SearchConfig = (SearchConfig)serializer.Deserialize(reader);
                            reader.Close();

                            _Dependencies["BaseConfig"] = new CacheDependency(path, DateTime.Now);
                        }
                    }
                }

                if (_DynamicSearchConfig != null)
                {
                    return _DynamicSearchConfig;
                }
                else
                {
                    _DynamicSearchConfig = new SearchConfig();
                }

                // Add dynamic fields here like categories
                string code = HttpContext.Current.Request["nc"];
                if (!String.IsNullOrEmpty(code))
                {
                    CatalogDto catalogs = CatalogContext.Current.GetCatalogDto(SiteContext.Current.SiteId);
                    foreach (CatalogDto.CatalogRow catalog in catalogs.Catalog)
                    {
#pragma warning disable CS0618 // Type or member is obsolete
                        CatalogNodes nodes = CatalogContext.Current.GetCatalogNodes(catalog.Name, code);
                        if (nodes.CatalogNode != null && nodes.CatalogNode.Length > 0)
                        {
                            List<SearchFilter> filters = new List<SearchFilter>();
                            foreach (CatalogNode nodeRow in nodes.CatalogNode)
#pragma warning restore CS0618 // Type or member is obsolete

                            {
                                SearchFilter configFilter = new SearchFilter();
                                configFilter.field = BaseCatalogIndexBuilder.FieldConstants.Node;

                                configFilter.Descriptions = new Descriptions();
                                configFilter.Descriptions.defaultLocale = "en";

                                Description desc = new Description();
                                desc.locale = "en";
                                desc.Value = new StoreResourceManager().GetString("CATEGORY_SHOP_LABEL");
                                configFilter.Descriptions.Description = new Description[] { desc };

                                SimpleValue val = new SimpleValue();
                                val.key = nodeRow.ID;
                                val.value = nodeRow.ID;
                                val.Descriptions = new Descriptions();
                                val.Descriptions.defaultLocale = "en";
                                Description desc2 = new Description();
                                desc2.locale = "en";
                                desc2.Value = nodeRow.ItemAttributes?["ItemDisplayName"]?.ToString() ?? nodeRow.Name;
                                val.Descriptions.Description = new Description[] { desc2 };

                                configFilter.Values = new SearchFilterValues();
                                configFilter.Values.SimpleValue = new SimpleValue[] { val };
                                filters.Add(configFilter);
                            }
                            _DynamicSearchConfig.SearchFilters = filters.ToArray();
                        }
                    }
                }

                List<SearchFilter> existingFilters = null;
                if (_DynamicSearchConfig.SearchFilters != null)
                    existingFilters = new List<SearchFilter>(_DynamicSearchConfig.SearchFilters);
                else
                    existingFilters = new List<SearchFilter>();

                foreach (SearchFilter filter in _SearchConfig.SearchFilters)
                {
                    existingFilters.Add(filter);
                }

                _DynamicSearchConfig.SearchFilters = existingFilters.ToArray();

                return _DynamicSearchConfig;
            }
        }


        /// <summary>
        /// The query string used for fetching search requests.
        /// </summary>
        /// <remarks>
        /// This will typically be the default value of HttpContext.Current.Request.QueryString.  However,
        /// if calling code needs to modify the collection, this can be used to inject a modifiable clone.
        /// </remarks>
        public NameValueCollection QueryString
        {
            get { return _queryString ?? HttpContext.Current.Request.QueryString; }
            set { _queryString = value; }
        }

        #endregion

        /// <summary>
        /// Determines whether [has config changed].
        /// </summary>
        /// <param name="name">The name.</param>
        /// <returns>
        /// 	<c>true</c> if [has config changed]; otherwise, <c>false</c>.
        /// </returns>
        private bool HasConfigChanged(string name)
        {
            lock (_LockObject)
            {
                if (!_Dependencies.ContainsKey(name))
                    return true;

                if (_Dependencies[name] == null || _Dependencies[name].HasChanged)
                    return true;
            }

            return false;
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SearchFilterHelper"/> class.
        /// </summary>
        public SearchFilterHelper()
        {
        }

        #region Search Methods


        /// <summary>
        /// Gets the suggestion keywords related to the current search.
        /// </summary>
        /// <returns></returns>
        public virtual string[] GetSuggestions()
        {
            NameValueCollection querystring = QueryString;
            return GetSuggestions(querystring["search"]);
        }

        /// <summary>
        /// Gets the suggestion keywords related to the search keyword passed.
        /// </summary>
        /// <param name="keyword">The keyword.</param>
        /// <returns></returns>
        public virtual string[] GetSuggestions(string keyword)
        {
            if (keyword == null)
                return null;

            if (this._Results == null)
                return null;

            //throw new NullReferenceException("Must perform search before calling for suggestions");

            string[] suggestions = _Results.Suggestions;
            return suggestions;
        }

        /// <summary>
        /// Gets the facets.
        /// </summary>
        /// <param name="cacheResults">if set to <c>true</c> [cache results].</param>
        /// <param name="cacheTimeout">The cache timeout.</param>
        /// <returns></returns>
        public virtual ISearchFacetGroup[] GetFacets(bool cacheResults, TimeSpan cacheTimeout)
        {
            var querystring = QueryString;
            var criteria = CreateSearchCriteria(querystring["search"], null);
            var readMode = cacheResults ? ReadMode.Cached : ReadMode.UnCached;

            var cacheKey = CatalogCache.CreateCacheKey("search-catalog-facets", criteria.CacheKey);
            var masterKeys = new[]
                {
                    MasterCacheKeys.AnyEntryKey,
                    MasterCacheKeys.AnyNodeKey,
                    MasterCacheKeys.AnyCatalogKey,
                    MasterCacheKeys.AnyEntryRelationKey,
                    MasterCacheKeys.AnyNodeRelationKey,
                    MasterCacheKeys.AnyCurrencyKey
                };

            return CatalogCache.ReadThrough(readMode, cacheKey, masterKeys, cacheTimeout, () => GetFacetsInternal(criteria));
        }

        private ISearchFacetGroup[] GetFacetsInternal(CatalogEntrySearchCriteria criteria)
        {
            return _Results != null ? _Results.FacetGroups : SearchEntries(criteria).FacetGroups;
        }

        /// <summary>
        /// Searches the entries.
        /// </summary>
        /// <returns></returns>
        public virtual ISearchResults SearchEntries()
        {
            NameValueCollection querystring = QueryString;
            return SearchEntries(querystring["search"], null);
        }

        /// <summary>
        /// Searches the entries.
        /// </summary>
        /// <param name="keywords">The keywords.</param>
        /// <param name="sort">The sort.</param>
        /// <param name="startIndex">The start index.</param>
        /// <param name="recordsToRetrieve">The records to retrieve.</param>
        /// <param name="count">The count.</param>
        /// <param name="responseGroup">The response group.</param>
        /// <param name="cacheResults">If set to <c>true</c> then fetch through cache, otherwise fetch
        /// fresh from db.</param>
        /// <param name="cacheTimeout">The cache timeout.</param>
        /// <returns></returns>
        [Obsolete("This method is no longer used. Use content types instead. Will remain at least until February 2019.")]
        public virtual Entries SearchEntries(string keywords, SearchSort sort, int startIndex, int recordsToRetrieve, out int count, CatalogEntryResponseGroup responseGroup, bool cacheResults, TimeSpan cacheTimeout)
        {
            CatalogEntrySearchCriteria criteria = CreateSearchCriteria(keywords, sort);
            criteria.StartingRecord = startIndex;
            criteria.RecordsToRetrieve = recordsToRetrieve;
            return SearchEntries(criteria, out count, responseGroup, cacheResults, cacheTimeout);
        }

        /// <summary>
        /// Searches the entries.
        /// </summary>
        /// <param name="criteria">The criteria.</param>
        /// <param name="count">The count.</param>
        /// <param name="responseGroup">The response group.</param>
        /// <param name="cacheResults">If set to <c>true</c> then fetch through cache, otherwise fetch
        /// fresh from db.</param>
        /// <param name="cacheTimeout">The cache timeout.</param>
        /// <returns>The search result.</returns>
        [Obsolete("This method is no longer used. Use content types instead. Will remain at least until February 2019.")]
        public virtual Entries SearchEntries(CatalogEntrySearchCriteria criteria, out int count, CatalogEntryResponseGroup responseGroup, bool cacheResults, TimeSpan cacheTimeout)
        {
            var readMode = cacheResults ? ReadMode.Cached : ReadMode.UnCached;

            // Assign new cache key, specific for site guid and response groups requested
            var cacheKey = CatalogCache.CreateCacheKey("seach-catalog-entries", responseGroup?.CacheKey, criteria.CacheKey);

            var masterKeys = new[]
                {
                    MasterCacheKeys.AnyEntryKey,
                    MasterCacheKeys.AnyCatalogKey,
                    MasterCacheKeys.AnyEntryRelationKey,
                    MasterCacheKeys.AnyNodeRelationKey,
                    MasterCacheKeys.AnyCurrencyKey
                };

            var entriesBoxed = CatalogCache.ReadThrough<object>(readMode, cacheKey, masterKeys, cacheTimeout,
                () => SearchEntriesInternal(criteria, responseGroup));

            if (entriesBoxed is Entries)
            {
                var cachedEntries = (Entries)entriesBoxed;
                count = cachedEntries.TotalResults;
                return cachedEntries;
            }

            count = 0;
            return new Entries { Entry = new Entry[0] };
        }

        [Obsolete("This method is no longer used. Use content types instead. Will remain at least until February 2019.")]
        private Entries SearchEntriesInternal(CatalogEntrySearchCriteria criteria, CatalogEntryResponseGroup responseGroup)
        {
            // Perform search with the criteria
            ISearchResults results = SearchEntries(criteria);

            // Get IDs of found Entries which we need
            int[] resultIndexes = results.GetKeyFieldValues<int>();

            // Retrieve actual entry objects from ECF database, with no caching
            // NOTE: because the GetCatalogEntries() cannot handle duplicated entry'id (violate unique constrain of Code when filling to Dataset)
            var entries = CatalogContext.Current.GetCatalogEntries(resultIndexes.Distinct().ToArray(), false, new TimeSpan(), responseGroup);
            entries.TotalResults = results.TotalCount;

            // Now add extra attribute called "_outline" which can be used to determine which category object belongs to
            int index = 0;
            if (entries.Entry != null)
            {
                foreach (Entry newEntry in entries.Entry)
                {
                    var attributes = new List<ItemAttribute>();
                    var metaAttributes = newEntry.ItemAttributes;

                    if (metaAttributes == null)
                    {
                        newEntry.ItemAttributes = new ItemAttributes();
                    }

                    if (metaAttributes != null && metaAttributes.Attribute != null)
                    {
                        attributes.AddRange(metaAttributes.Attribute);
                    }

                    var vals = results.GetFieldValue<string>(index, BaseCatalogIndexBuilder.FieldConstants.Outline);
                    if (vals != null)
                    {
                        attributes.Add(ObjectHelper.CreateAttribute(BaseCatalogIndexBuilder.FieldConstants.Outline, "Outline",
                            "string[]", vals));
                    }

                    newEntry.ItemAttributes.Attribute = attributes.ToArray();

                    index++;
                }
            }
            return entries;
        }

        /// <summary>
        /// Searches the entries.
        /// </summary>
        /// <param name="keywords">The keywords.</param>
        /// <param name="sort">The sort.</param>
        /// <returns></returns>
        public virtual ISearchResults SearchEntries(string keywords, SearchSort sort)
        {
            CatalogEntrySearchCriteria criteria = CreateSearchCriteria(keywords, sort);
            return SearchEntries(criteria);
        }

        /// <summary>
        /// Searches the entries.
        /// </summary>
        /// <param name="criteria">The criteria.</param>
        /// <returns></returns>
        public virtual ISearchResults SearchEntries(CatalogEntrySearchCriteria criteria)
        {
            try
            {
                _Results = Manager.Search(criteria);
            }
            catch (SystemException)
            {
                if (HttpContext.Current.IsDebuggingEnabled)
                    throw;
            }

            // Perform fuzzy search if nothing has been found
            if (_Results.TotalCount == 0)
            {
                criteria.IsFuzzySearch = true;
                criteria.FuzzyMinSimilarity = 0.7f;
                _Results = Manager.Search(criteria);
            }

            return _Results;
        }

        public virtual ISearchResults SearchEntries(CatalogEntrySearchCriteria criteria, bool isFuzzySearch, float minSimilarity)
        {
            try
            {
                criteria.IsFuzzySearch = isFuzzySearch;
                criteria.FuzzyMinSimilarity = minSimilarity;
                _Results = Manager.Search(criteria);
            }
            catch (SystemException)
            {
                if (HttpContext.Current.IsDebuggingEnabled)
                    throw;
            }
            return _Results;
        }

        /// <summary>
        /// Creates the search criteria.
        /// </summary>
        /// <param name="keywords">The keywords.</param>
        /// <param name="sort">The sort.</param>
        /// <returns></returns>
        public virtual CatalogEntrySearchCriteria CreateSearchCriteria(string keywords, SearchSort sort)
        {
            NameValueCollection querystring = QueryString;
            string currentNodeCode = querystring["nc"];
            CatalogEntrySearchCriteria criteria = new CatalogEntrySearchCriteria();
            criteria.Currency = SiteContext.Current.Currency.CurrencyCode;

            if (!String.IsNullOrEmpty(currentNodeCode))
            {
                // Only populate when count is 0 to allow overriding how outlines are calculated
                if (criteria.Outlines.Count == 0)
                    criteria.Outlines = GetOutlinesForNode(currentNodeCode);
            }

            criteria.SearchPhrase = keywords;
            criteria.SearchIndex.Add(querystring["filter"]);
            criteria.Sort = sort;

            // Add all filters
            foreach (SearchFilter filter in SearchConfig.SearchFilters)
            {
                // Check if we already filtering
                if (querystring.Get(filter.field) != null)
                    continue;

                criteria.Add(filter);
            }

            // Get selected filters
            SelectedFilter[] filters = SelectedFilters;
            if (filters.Length != 0)
            {
                foreach (SelectedFilter filter in filters)
                {
                    if (filter.PriceRangeValue != null)
                        criteria.Add(filter.Filter.field, filter.PriceRangeValue);
                    if (filter.RangeValue != null)
                        criteria.Add(filter.Filter.field, filter.RangeValue);
                    if (filter.SimpleValue != null)
                        criteria.Add(filter.Filter.field, filter.SimpleValue);
                }
            }

            return criteria;
        }

        #region Private Helper Methods
        /// <summary>
        /// Gets the outlines for node.
        /// </summary>
        /// <param name="nodeCode">The node code.</param>
        /// <returns></returns>
        public static StringCollection GetOutlinesForNode(string nodeCode)
        {
            string path = nodeCode;
            StringCollection col = new StringCollection();

            var id = NodeIdentityResolver.Service.GetIdsByCodes(new[] {nodeCode}).Values.FirstOrDefault();
            if (id == 0)
            {
                return col;
            }
            var catalogNode = CatalogContext.Current.GetCatalogNodeDto(id);
            CatalogDto catalog = CatalogContext.Current.GetCatalogDto(catalogNode.CatalogNode[0].CatalogId);
            int catalogNodeId = catalogNode.CatalogNode[0].ParentNodeId;
            if (catalogNodeId != 0)
            {
                while ((catalogNode = CatalogContext.Current.GetCatalogNodeDto(catalogNodeId)) != null)
                {
                    if (catalogNode.CatalogNode.Any())
                    {
                        break;
                    }

                    path = String.Format("{1}/{0}", path, catalogNode.CatalogNode[0].Code);

                    catalogNodeId = catalogNode.CatalogNode[0].ParentNodeId;
                    if (catalogNodeId == 0)
                    {
                        break;
                    }
                }
            }

            path = $"{catalog.Catalog[0].Name}/{path}";

            col.Add($"{path}*");
            return col;
        }

        /// <summary>
        /// Gets the outlines for node.
        /// </summary>
        /// <param name="node">The node.</param>
        /// <returns></returns>
        [Obsolete("This method is no longer used. Use GetOutlinesForNode(string nodeCode) instead. Will remain at least until February 2019.")]
        public static StringCollection GetOutlinesForNode(CatalogNode node)
        {
            return GetOutlinesForNode(node.ID);
        }
        #endregion
        #endregion

        #region QueryString Methods
        /// <summary>
        /// Gets the clean URL without filters.
        /// </summary>
        /// <returns></returns>
        public string GetCleanUrl()
        {
            string url = CurrentUrl;

            foreach (SelectedFilter filter in this.SelectedFilters)
            {
                url = GetQueryStringNavigateUrl(url, filter.Filter.field, String.Empty, true);
            }

            return url;
        }

        /// <summary>
        /// Gets the query string navigate URL.
        /// </summary>
        /// <param name="field">The field.</param>
        /// <param name="key">The key.</param>
        /// <returns></returns>
        public static string GetQueryStringNavigateUrl(string field, string key)
        {
            return GetQueryStringNavigateUrl(field, key, false);
        }

        /// <summary>
        /// Gets the query string navigate URL.
        /// </summary>
        /// <param name="field">The field.</param>
        /// <param name="remove">if set to <c>true</c> [remove].</param>
        /// <returns></returns>
        public static string GetQueryStringNavigateUrl(string field, bool remove)
        {
            return GetQueryStringNavigateUrl(field, String.Empty, remove);
        }

        /// <summary>
        /// Gets the query string navigate URL.
        /// </summary>
        /// <param name="field">The field.</param>
        /// <param name="key">The key.</param>
        /// <param name="remove">if set to <c>true</c> [remove].</param>
        /// <returns></returns>
        public static string GetQueryStringNavigateUrl(string field, string key, bool remove)
        {
            return GetQueryStringNavigateUrl(CurrentUrl, field, key, remove);
        }

        /// <summary>
        /// Gets the query string navigate URL.
        /// </summary>
        /// <param name="url">The URL.</param>
        /// <param name="field">The field.</param>
        /// <param name="key">The key.</param>
        /// <param name="remove">if set to <c>true</c> [remove].</param>
        /// <returns></returns>
        public static string GetQueryStringNavigateUrl(string url, string field, string key, bool remove)
        {
            string queryField = field;

            StringBuilder builder = new StringBuilder();

            string path = String.Empty;
            if (url.Contains("?"))
            {
                builder.Append(url.Substring(0, url.IndexOf("?")));
                builder.Append("?");

                NameValueCollection query = new NameValueCollection();
                string queryString = url.Substring(url.IndexOf("?") + 1);
                string[] queryParamsArray = queryString.Split(new char[] { '&' });
                foreach (string queryParam in queryParamsArray)
                {
                    string[] queryParamArray = queryParam.Split(new char[] { '=' });
                    if (!queryParamArray[0].Equals(queryField, StringComparison.OrdinalIgnoreCase))
                    {
                        builder.Append(queryParamArray[0]);
                        builder.Append("=");
                        if (queryParamArray.Length > 1)
                            builder.Append(queryParamArray[1]);
                        builder.Append("&");
                    }
                }
            }
            else
            {
                builder.Append(url);
                builder.Append("?");
            }

            if (!remove)
            {
                builder.Append(queryField);
                builder.Append("=");
                builder.Append(key);
            }
            else
            {
                builder.Remove(builder.Length - 1, 1);
            }


            return builder.ToString();
        }

        SelectedFilter[] _SelectedFilters = null;
        /// <summary>
        /// Gets the selected filters.
        /// </summary>
        /// <value>The selected filters.</value>
        public virtual SelectedFilter[] SelectedFilters
        {
            get
            {
                if (_SelectedFilters != null)
                    return _SelectedFilters;

                NameValueCollection querystring = QueryString;
                List<SelectedFilter> filters = new List<SelectedFilter>();

                // Cycle through query string to see if we have any filters specified
                foreach (string key in querystring.AllKeys)
                {
                    bool foundTerm = false;
                    SearchConfig config = SearchConfig;
                    foreach (SearchFilter filter in config.SearchFilters)
                    {
                        // we found a filter!
                        if (filter.field.Equals(key, StringComparison.OrdinalIgnoreCase))
                        {
                            // Now find the value
                            string val = querystring[key];

                            // Simple values
                            if (filter.Values.SimpleValue != null)
                            {
                                foreach (SimpleValue value in filter.Values.SimpleValue)
                                {
                                    if (value.key.Equals(val, StringComparison.OrdinalIgnoreCase))
                                    {
                                        filters.Add(new SelectedFilter(filter, value));
                                        foundTerm = true;
                                        break;
                                    }
                                }
                            }

                            // Range values
                            if (filter.Values.RangeValue != null)
                            {
                                foreach (RangeValue value in filter.Values.RangeValue)
                                {
                                    if (value.key.Equals(val, StringComparison.OrdinalIgnoreCase))
                                    {
                                        filters.Add(new SelectedFilter(filter, value));
                                        foundTerm = true;
                                        break;
                                    }
                                }
                            }

                            // Price Range values
                            if (filter.Values.PriceRangeValue != null)
                            {
                                foreach (PriceRangeValue value in filter.Values.PriceRangeValue)
                                {
                                    if (value.key.Equals(val, StringComparison.OrdinalIgnoreCase) && value.currency.Equals(SiteContext.Current.Currency.CurrencyCode, StringComparison.OrdinalIgnoreCase))
                                    {
                                        filters.Add(new SelectedFilter(filter, value));
                                        foundTerm = true;
                                        break;
                                    }
                                }
                            }

                        }

                        if (foundTerm)
                            break;
                    }
                }

                _SelectedFilters = filters.ToArray();
                return _SelectedFilters;
            }
        }
        #endregion
    }
}
